/*
 * Copyright 2016  Veselin Georgiev,
 * anrieffNOSPAM @ mgail_DOT.com (convert to gmail)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * This file contains a list of internal codes we use in detection. It is
 * of no external use and isn't a complete list of intel products.
 */
	CODE2(PENTIUM, 2000),
	
	CODE(IRWIN),
	CODE(POTOMAC),
	CODE(GAINESTOWN),
	CODE(WESTMERE),
	
	CODE(PENTIUM_M),
	CODE(NOT_CELERON),	
	
	CODE(CORE_SOLO),
	CODE(MOBILE_CORE_SOLO),
	CODE(CORE_DUO),
	CODE(MOBILE_CORE_DUO),
	
	CODE(WOLFDALE),
	CODE(MEROM),
	CODE(PENRYN),
	CODE(QUAD_CORE),
	CODE(DUAL_CORE_HT),
	CODE(QUAD_CORE_HT),
	CODE(MORE_THAN_QUADCORE),
	CODE(PENTIUM_D),
	
	CODE(SILVERTHORNE),
	CODE(DIAMONDVILLE),
	CODE(PINEVIEW),
	CODE(CEDARVIEW),
